package com.suse.tracecompass.cpugraph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import org.eclipse.swt.graphics.*;
import org.eclipse.linuxtools.internal.lttng2.kernel.core.Attributes;
import org.eclipse.linuxtools.internal.lttng2.kernel.core.StateValues;
import org.eclipse.linuxtools.lttng2.kernel.core.trace.LttngKernelTrace;
import org.eclipse.linuxtools.tmf.core.exceptions.*;
import org.eclipse.linuxtools.tmf.core.interval.ITmfStateInterval;
import org.eclipse.linuxtools.tmf.core.statesystem.ITmfStateSystem;
import org.eclipse.linuxtools.tmf.ui.widgets.timegraph.StateItem;
import org.eclipse.linuxtools.tmf.ui.widgets.timegraph.TimeGraphPresentationProvider;
import org.eclipse.linuxtools.tmf.ui.widgets.timegraph.TimeGraphViewer;
import org.eclipse.linuxtools.tmf.ui.widgets.timegraph.model.ITimeEvent;

@SuppressWarnings("restriction")
public class CpuGraphPresentationProvider extends TimeGraphPresentationProvider {
    private static final int MAX_CPUS = 128;
    private static final int STATES = 5;
    private RGB[] cpu_rgb = new RGB[MAX_CPUS];
    private Color[] cpu_colors = new Color[MAX_CPUS];
    private final Random random = new Random(0xf00d);
    private static final float GOLDEN_RATIO_CONJUGATE = 0.618033988749895f;
    /*
     * kudos to http://martin.ankerl.com/2009/12/09/how-to-create-random-colors-programmatically/
     */

    private boolean zoomed = false;

    private TimeGraphViewer viewer;

    private List<CpuGraphEvent> eventsSeen = new ArrayList<>();

    @Override
    public String getEventName(ITimeEvent event) {
        if (!(event instanceof CpuGraphEvent))
            return "other ???";
        return getStateName(((CpuGraphEvent) event).state);
    }

    private String getStateName(int state) {
        switch (state) {
        case StateValues.CPU_STATUS_IDLE:
            return "idle";
        case StateValues.CPU_STATUS_IRQ:
            return "IRQ";
        case StateValues.CPU_STATUS_SOFTIRQ:
            return "softIRQ";
        case StateValues.CPU_STATUS_RUN_SYSCALL:
            return "syscall";
        case StateValues.CPU_STATUS_RUN_USERMODE:
            return "userspace";
        }
        return "other ???";
    }

    private StateItem[] stateTable = null;

    @Override
    public StateItem[] getStateTable() {
        if (stateTable != null)
            return stateTable;

        ArrayList<StateItem> states = new ArrayList<StateItem>((MAX_CPUS * STATES) + 1);
        for (int i = 0; i < MAX_CPUS; i++) {
            String cpustr = "CPU " + i + " ";
            float hue = (random.nextFloat() + GOLDEN_RATIO_CONJUGATE) % 1;
            hue *= 360;

            cpu_rgb[i] = new RGB(hue, 0.95f, 0.5f);
            // idle
            states.add(new StateItem(new RGB(hue, 0.8f, 0.90f), cpustr + getStateName(StateValues.CPU_STATUS_IDLE)));
            // userspace + syscall
            states.add(new StateItem(new RGB(hue, 0.95f, 0.5f), cpustr + getStateName(StateValues.CPU_STATUS_RUN_USERMODE)));
            states.add(new StateItem(new RGB(hue, 0.95f, 0.6f), cpustr + getStateName(StateValues.CPU_STATUS_RUN_SYSCALL)));
            // irq
            states.add(new StateItem(new RGB(hue, 0.6f, 0.7f), cpustr + getStateName(StateValues.CPU_STATUS_IRQ)));
            states.add(new StateItem(new RGB((hue + 4) % 360, 0.7f, 0.75f), cpustr + getStateName(StateValues.CPU_STATUS_SOFTIRQ)));
        }

        states.add(new StateItem(new RGB(200, 200, 200), getStateName(-1))); // unknown

        StateItem[] array = new StateItem[states.size()];
        stateTable = states.toArray(array);
        return stateTable;
    }

    @Override
    public int getStateTableIndex(ITimeEvent event) {
        if (!(event instanceof CpuGraphEvent))
            return MAX_CPUS * STATES;

        CpuGraphEvent e = (CpuGraphEvent) event;
        if (e.getEntry() instanceof ProcessEntry)
            return e.cpuid * STATES + 2;
        else
            return e.cpuid * STATES + e.state;
    }

    private void drawlink(CpuGraphEvent e1, CpuGraphEvent e2, GC gc, Rectangle bounds) {
        // leading/trailing events of scan - nothing to do here
        if (e1 == null || e2 == null)
            return;
        // on same line
        if (e1.y == e2.y)
            return;
        
        if (!zoomed) {
        	return;
        }

        Color color = cpu_colors[e1.cpuid];
        if (color == null) {
            color = new Color(gc.getDevice(), cpu_rgb[e1.cpuid]);
            cpu_colors[e1.cpuid] = color;
        }
        gc.setForeground(color);

        int y1, y2;
        // find the top Y
        if (e1.y > e2.y) {
            y1 = e2.y;
            y2 = e1.y + e1.h - 1;
        } else {
            y1 = e1.y;
            y2 = e2.y + e2.h - 1;
        }

        int x = e1.iter == -1 ? e2.x1 : e1.x2;

        gc.drawLine(x, y1, x, y2);
    }

    @Override
    public void postDrawEvent(ITimeEvent event, Rectangle bounds, GC gc) {
        if (!(event instanceof CpuGraphEvent)) {
        	return;
        }
        
        CpuGraphEvent ev = (CpuGraphEvent) event;
        ev.y = bounds.y;
        ev.h = bounds.height;
        ev.x1 = bounds.x;
        ev.x2 = bounds.x + bounds.width;

        /*
         * ev.iter++; if (isZoomed()) { drawlink(ev.prev, ev, bounds.height,
         * gc); drawlink(ev, ev.next, bounds.height, gc); }
         */
        eventsSeen.add(ev);
    }

    @Override
    public void postDrawControl(Rectangle bounds, GC gc) {

        // draw links
        for (CpuGraphEvent ev : eventsSeen) {
            if (ev.prev != null)
                ev.prev.iter = -1;
            if (ev.next != null)
                ev.next.iter = -1;
        }
        for (CpuGraphEvent ev : eventsSeen) {
            ev.iter = 1;
        }

        for (CpuGraphEvent ev : eventsSeen) {
            if (ev.prev != null && ev.prev.iter == -1) {
                if (ev.prev.getEntry() instanceof ProcessEntry)
                    ev.prev.y = bounds.y;
                else
                    ev.prev.y = bounds.y + bounds.height;
            }
            drawlink(ev.prev, ev, gc, bounds);

            if (ev.next != null && ev.next.iter == -1) {
                if (ev.next.getEntry() instanceof ProcessEntry)
                    ev.next.y = bounds.y;
                else
                    ev.next.y = bounds.y + bounds.height;
            }
            drawlink(ev, ev.next, gc, bounds);
        }

        // clear seen list
        eventsSeen.clear();

/*        if (selectedTime == -1)
            return;
        long time0 = viewer.getTime0();
        long time1 = viewer.getTime1();
        double pixelsPerNanoSec = (bounds.width <= 1) ? 0 : (double) (bounds.width - 1) / (time1 - time0);
        int x = bounds.x + (int) ((double) (selectedTime - time0) * pixelsPerNanoSec);
        if (x >= 0 && x < bounds.x + bounds.width) {
            if (time_color == null)
                time_color = new Color(gc.getDevice(), new RGB(255, 0, 0));
            gc.setForeground(time_color);
            gc.drawLine(x, bounds.y, x, bounds.y + bounds.height);
        }*/
    }

    @Override
    public Map<String, String> getEventHoverToolTipInfo(ITimeEvent event) {
        Map<String, String> tooltip = new HashMap<String, String>();

        if (event instanceof CpuGraphEvent) {
            CpuGraphEvent ev = (CpuGraphEvent) event;
            tooltip.put("CPU", String.valueOf(ev.cpuid));

            if (ev.getEntry() instanceof ProcessEntry) {
                ProcessEntry entry = (ProcessEntry) ev.getEntry();
                tooltip.put("TID", String.valueOf(entry.pid));
            }
            if (ev.getEntry() instanceof CpuEntry) {
                CpuEntry entry = (CpuEntry) ev.getEntry();
                ITmfStateSystem ssq = entry.getTrace().getStateSystems().get(LttngKernelTrace.STATE_ID);

                try {
                    List<ITmfStateInterval> fullState = ssq.queryFullState(event.getTime());
                    int tidQuark = ssq.getQuarkAbsolute(Attributes.CPUS, String.valueOf(entry.cpuid), Attributes.CURRENT_THREAD);
                    int tid = fullState.get(tidQuark).getStateValue().unboxInt();
                    tooltip.put("TID", String.valueOf(tid));
                    int execnameQuark = ssq.getQuarkAbsolute(Attributes.THREADS, String.valueOf(tid), Attributes.EXEC_NAME);
                    if (tid != 0)
                        tooltip.put("Process name", fullState.get(execnameQuark).getStateValue().unboxStr());

                    int syscallQuark = ssq.getQuarkAbsolute(Attributes.THREADS, String.valueOf(tid), Attributes.SYSTEM_CALL);
                    if (ev.state == StateValues.CPU_STATUS_RUN_SYSCALL) {
                        tooltip.put("Syscall", fullState.get(syscallQuark).getStateValue().unboxStr());
                    }
                } catch (StateSystemDisposedException e) {
                    e.printStackTrace();
                } catch (AttributeNotFoundException e) {
                    e.printStackTrace();
                } catch (TimeRangeException e) {
                    e.printStackTrace();
                } catch (StateValueTypeException e) {
                    e.printStackTrace();
                }
            }
        }

        return tooltip;
    }

    boolean isZoomed() {
        return zoomed;
    }

    void setZoomed(boolean zoomed) {
        this.zoomed = zoomed;
    }

    public TimeGraphViewer getViewer() {
        return viewer;
    }

    public void setViewer(TimeGraphViewer viewer) {
        this.viewer = viewer;
    }
}
