package com.suse.tracecompass.cpugraph;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.TreeMap;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.linuxtools.internal.lttng2.kernel.core.Attributes;
import org.eclipse.linuxtools.internal.lttng2.kernel.core.StateValues;
import org.eclipse.linuxtools.lttng2.kernel.core.trace.LttngKernelTrace;
import org.eclipse.linuxtools.tmf.core.exceptions.*;
import org.eclipse.linuxtools.tmf.core.interval.ITmfStateInterval;
import org.eclipse.linuxtools.tmf.core.statesystem.ITmfStateSystem;
import org.eclipse.linuxtools.tmf.core.statevalue.ITmfStateValue;
import org.eclipse.linuxtools.tmf.core.trace.ITmfTrace;
import org.eclipse.linuxtools.tmf.core.trace.TmfTraceManager;
import org.eclipse.linuxtools.tmf.ui.views.timegraph.AbstractTimeGraphView;
import org.eclipse.linuxtools.tmf.ui.widgets.timegraph.model.ITimeEvent;
import org.eclipse.linuxtools.tmf.ui.widgets.timegraph.model.TimeGraphEntry;

@SuppressWarnings("restriction")
public class CpuGraphView extends AbstractTimeGraphView {

    public static final String ID = "com.suse.tracecompass.cpugraph.view"; //$NON-NLS-1$

    private static final long ZOOM_LIMIT = 5000;

    private static final String[] COLUMN_NAMES = new String[] { "Process", "TID" };
    private static final String[] FILTER_COLUMN_NAMES = new String[] { "Process", "TID" };
    
    // weights of the treeview pane vs the gantt chart pane
    private static final int[] PANE_WEIGHTS = new int[] { 1, 4 };

    private static final DummyEntry DUMMY_ENTRY = new DummyEntry("", Long.MIN_VALUE, Long.MAX_VALUE);

    private final CpuGraphPresentationProvider provider;
    
    private long renderedRangeStart = Long.MAX_VALUE;
    private long renderedRangeEnd = Long.MIN_VALUE;
    private long renderedResolution = -1;
    
    public CpuGraphView() {
    	this(new CpuGraphPresentationProvider());
    }

    public CpuGraphView(CpuGraphPresentationProvider provider) {
        super(ID, provider);
        this.provider = provider;

        setTreeColumns(COLUMN_NAMES);
        setFilterColumns(FILTER_COLUMN_NAMES);
        TreeLabelProvider labelProvider = new CpuGraphTreeLabelProvider();
        setTreeLabelProvider(labelProvider);
        setWeight(PANE_WEIGHTS);
        // setEntryComparator(something);
    }

    private ArrayList<CpuEntry> cpus = new ArrayList<CpuEntry>();

    // not a set because we need to retrieve the instances that are within
    private TreeMap<ProcessEntry, ProcessEntry> processes = new TreeMap<ProcessEntry, ProcessEntry>();

    protected static class CpuGraphTreeLabelProvider extends TreeLabelProvider {

        @Override
        public String getColumnText(Object element, int columnIndex) {
            if (element instanceof DummyEntry)
                return "";
            CpuGraphEntry entry = (CpuGraphEntry) element;

            switch (columnIndex) {
            case 0:
                return entry.getName();
            case 1:
                if (entry instanceof ProcessEntry) {
                    return String.valueOf(((ProcessEntry)entry).pid);
                }
            default:
                return "";
            }
        }

    }

    // ------------------------------------------------------------------------
    // ViewPart
    // ------------------------------------------------------------------------

    @Override
    public void createPartControl(Composite parent) {
        super.createPartControl(parent);
        provider.setViewer(getTimeGraphViewer());

        /*
        viewer.addFilter(new EmptyRangeFilter());
        installToolbars();
        */
    }

    private CpuEntry getCpuEntry(LttngKernelTrace trace, int cpuid) {
        if (cpuid >= cpus.size()) {
            for (int i = cpus.size(); i <= cpuid; i++)
                cpus.add(new CpuEntry("CPU " + String.valueOf(i), trace, i));
        }
        return cpus.get(cpuid);
    }

    private ProcessEntry getProcessEntry(LttngKernelTrace trace, int pid, String name) {
        ProcessEntry pe = new ProcessEntry(name, trace, pid);
        ProcessEntry right = processes.get(pe);
        if (right == null) {
            processes.put(pe, pe);
            return pe;
        } else {
            return right;
        }
    }

    @Override
    protected void buildEventList(ITmfTrace rootTrace, IProgressMonitor monitor) {
        monitor.beginTask("experiment selected", IProgressMonitor.UNKNOWN);
        cpus.clear();
        processes.clear();

        setStartTime(Long.MAX_VALUE);
        setEndTime(Long.MIN_VALUE);
        
        /* this code assumes that only one LTTNG trace is present. fails if it isn't the case */
        /* to fix this, we would have to keep per-trace entry lists or something */
        LttngKernelTrace lastSeenKernelTrace = null;

        for (ITmfTrace candidateTrace : TmfTraceManager.getTraceSet(rootTrace)) {
            if (monitor.isCanceled()) {
                return;
            }
            if (!(candidateTrace instanceof LttngKernelTrace)) {
                continue;
            }

            LttngKernelTrace trace = (LttngKernelTrace) candidateTrace;
            lastSeenKernelTrace = trace;
            ITmfStateSystem ssq = trace.getStateSystems().get(LttngKernelTrace.STATE_ID);
            if (!ssq.waitUntilBuilt()) {
                return;
            }

            long start = ssq.getStartTime();
            long end = ssq.getCurrentEndTime() + 1;
            setStartTime(Math.min(getStartTime(), start));
            setEndTime(Math.max(getEndTime(), end));

            /* enumerate CPUs, don't bother cancelling, this one is quick */
            monitor.subTask("enumerating CPUs");
            List<Integer> cpuQuarks = ssq.getQuarks(Attributes.CPUS, "*"); //$NON-NLS-1$
            for (int cpuQuark : cpuQuarks) {
                int cpuid = Integer.valueOf(ssq.getAttributeName(cpuQuark));
                getCpuEntry(trace, cpuid);
                monitor.worked(1);
            }

            /* enumerate processes */
            monitor.subTask("enumerating pids");
            enumerateProcesses(trace, monitor, ssq, start, end);
        }
        
        List<TimeGraphEntry> entryList = buildEntryList();
        for (ITmfTrace aTrace : TmfTraceManager.getTraceSet(rootTrace)) {
        	putEntryList(aTrace, entryList);
        	if (aTrace.equals(getTrace())) {
        		refresh();
        	}
        }

        monitor.subTask("building view");
        /*
         * setViewStartTime(start); setViewEndTime(end);
         * setDefaultInput(getRelevantEntries());
         */
        long resolution = Math.max(1, (getEndTime() - getStartTime()) / getDisplayWidth());
        renderRange(lastSeenKernelTrace, getStartTime(), getEndTime(), resolution, monitor);
    }
    
    private void enumerateProcesses(LttngKernelTrace trace, IProgressMonitor monitor, ITmfStateSystem ssq, long start, long end) {
        List<Integer> tQuarks = ssq.getQuarks(Attributes.THREADS, "*"); //$NON-NLS-1$
        for (int tQuark : tQuarks) {
            if (monitor.isCanceled()) return;

            int tid;
            try {
                tid = Integer.valueOf(ssq.getAttributeName(tQuark));
                // skip idle tid
                // NOTE: do it, because enumerating states of idle TID takes an insane amount of time
                // (and you get nothing meaningful anyway)
                if (tid == 0) continue;
            } catch (NumberFormatException e) {
                continue;
            }
            try {
                int execname = ssq.getQuarkRelative(tQuark, Attributes.EXEC_NAME);
                List<ITmfStateInterval> execNameIntervals = ssq.queryHistoryRange(execname, start, end);

                String prevname = null;
                for (ITmfStateInterval si : execNameIntervals) {
                    if (monitor.isCanceled()) return;
                    if (!si.getStateValue().isNull() && si.getStateValue().getType() == ITmfStateValue.Type.STRING) {
                        String name = si.getStateValue().unboxStr();
                        if (!name.equals(prevname)) {
                            getProcessEntry(trace, tid, name);
                        }
                        prevname = name;
                    }
                }
                monitor.worked(1);

            } catch (AttributeNotFoundException | TimeRangeException | StateValueTypeException e) {
                continue;
            } catch (StateSystemDisposedException e) {
                return;
            }
        }
    }

    private List<TimeGraphEntry> buildEntryList() {

        List<TimeGraphEntry> input = new ArrayList<>();
        for (ProcessEntry e : processes.navigableKeySet())
            input.add(e);
        for (int i = 0; i < 5; i++)
            input.add(DUMMY_ENTRY);
        for (CpuEntry e : cpus)
            input.add(e);

        return input;
    }   

    @Override
    protected List<ITimeEvent> getEventList(TimeGraphEntry tgentry, long startTime, long endTime, long resolution, IProgressMonitor monitor) {
        if (!(tgentry instanceof CpuGraphEntry)) {
            return null;
        }
        
        CpuGraphEntry entry = (CpuGraphEntry) tgentry;
        renderRange(entry.getTrace(), startTime, endTime - 1, resolution, monitor);
        
        final long realStart = Math.max(startTime, entry.getStartTime());
        final long realEnd = Math.min(endTime, entry.getEndTime());
        if (realEnd <= realStart) {
            return null;
        }

        return entry.getEventBuffer();
    }
    
    private void renderRange(LttngKernelTrace trace, long start, long end, long resolution, IProgressMonitor monitor)
    {
        if (start == renderedRangeStart && end == renderedRangeEnd
            && resolution == renderedResolution) {
            return;
        }
        
        ITmfStateSystem ssq = trace.getStateSystems().get(LttngKernelTrace.STATE_ID);
        start = Math.max(start,  ssq.getStartTime());
        end = Math.min(end, ssq.getCurrentEndTime() + 1);
        
        if (start >= end) {
            return;
        }
        
        renderedRangeStart = start;
        renderedRangeEnd = end;
        renderedResolution = resolution;
                
        monitor.beginTask("rendering range", IProgressMonitor.UNKNOWN);
        
        /* clear out event buffers */
        for (CpuEntry entry : cpus) {
        	entry.getEventBuffer().clear();
        }
        for (ProcessEntry entry : processes.navigableKeySet()) {
        	entry.getEventBuffer().clear();
        }

        for (CpuEntry entry : cpus) {
            monitor.subTask("processing CPU " + entry.cpuid);
            CpuGraphEvent prev = null;
            CpuGraphEvent event = null;
            try {
                int cpuQuark = ssq.getQuarkAbsolute(Attributes.CPUS, String.valueOf(entry.cpuid));
                int tidQuark = ssq.getQuarkRelative(cpuQuark, Attributes.CURRENT_THREAD);
                int stateQuark = ssq.getQuarkRelative(cpuQuark, Attributes.STATUS);

                if (monitor.isCanceled())
                    return;

                // list all intervals in specified range
                List<ITmfStateInterval> intervals = ssq.queryHistoryRange(stateQuark, start, end, resolution, monitor);

                for (ITmfStateInterval interval : intervals) {
                    if (monitor.isCanceled())
                        return;

                    if (interval.getStateValue().isNull() || interval.getStateValue().getType() != ITmfStateValue.Type.INTEGER)
                        continue;

                    int status = interval.getStateValue().unboxInt();

                    long startTime = interval.getStartTime();
                    long endTime = interval.getEndTime();

                    if (status != StateValues.CPU_STATUS_RUN_USERMODE) {
                        event = new CpuGraphEvent(entry, startTime, endTime - startTime, status, entry.cpuid);
                        entry.getEventBuffer().add(event);
                    } else {
                        // if (status == StateValues.CPU_STATUS_RUN_USERMODE) {
                        int tid = ssq.querySingleState(startTime, tidQuark).getStateValue().unboxInt();
                        int nameQuark = ssq.getQuarkAbsolute(Attributes.THREADS, String.valueOf(tid), Attributes.EXEC_NAME);
                        String name = ssq.querySingleState(startTime, nameQuark).getStateValue().unboxStr();
                        ProcessEntry pe = getProcessEntry(trace, tid, name);
                        event = new CpuGraphEvent(pe, startTime, endTime - startTime, status, entry.cpuid);
                        pe.getEventBuffer().add(event);
                    }

                    // TODO fill in unknowns if the events don't follow each
                    // other

                    if (prev != null)
                        prev.next = event;
                    event.prev = prev;
                    prev = event;
                    monitor.worked(1);
                }
            } catch (AttributeNotFoundException e) {
                System.out.println("attr not found for cpu: " + entry.cpuid);
                e.printStackTrace();
            } catch (TimeRangeException | StateValueTypeException e) {
                e.printStackTrace();
            } catch (StateSystemDisposedException e) {
                break;
            }

            List<ITimeEvent> bufcopy = new ArrayList<ITimeEvent>(entry.getEventBuffer());
            entry.setEventList(bufcopy);

            for (ProcessEntry pe : processes.navigableKeySet()) {
                bufcopy = new ArrayList<ITimeEvent>(pe.getEventBuffer());
                Collections.sort(bufcopy, CpuGraphEvent.COMPARATOR);
                pe.setEventList(bufcopy);
            }

            // invalidate();
            monitor.worked(1);
        }

        provider.setZoomed(resolution < ZOOM_LIMIT);
        //if (trace.equals(getTrace())) { redraw(); }
        redraw();
    }

}
